#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# opsidirectoryconnector complements the desktop management solution opsi
# (open pc server integration) http://www.opsi.org
# Copyright (C) 2016-2018 uib GmbH <info@uib.de>

"""
opsi directory connector

Connects LDAP-compatible directories and opsi.

Please visit http://uib.de/ for more information.
"""

import argparse
import logging
import json
import sys
import time
from collections import namedtuple
from contextlib import contextmanager
from functools import partial
from urllib.parse import urlsplit

import requests

from ldap3 import Connection, Server
try:
    from ldap3 import LDAPKeyError
except ImportError:
    # Probably ldap3 > 2.0
    from ldap3.core.exceptions import LDAPKeyError

try:
    from ldap3 import SUBTREE
except ImportError:
    SUBTREE = None

__version__ = '26'

USERAGENT = "opsi directory connector v{0}".format(__version__)
DEFAULT_ATTRIBUTES = set(['cn', 'objectCategory', 'objectClass'])
NAME_ADDITION = '_'

LOGGER = logging.getLogger('opsidirectoryconnector')

ClientMapping = namedtuple("ClientMapping", "id description notes hardwareAddress ipAddress inventoryNumber oneTimePassword")
GroupMapping = namedtuple("GroupMapping", "id description notes")


class ModulesError(Exception):
    pass


def runConnector():
    cliArguments = parseCommandlineArguments()
    configureLogger(LOGGER, level=cliArguments.logLevel * 10)
    config = readConfig(cliArguments.config)
    config = convertConfig(config)
    rootGroup = config['behaviour']['root_dir_in_opsi']
    LOGGER.debug("root group: {}".format(rootGroup))

    if not rootGroup:
        raise ValueError("Invalid configuration for behaviour -> root_dir_in_opsi")

    # Remove the value before connecting to avoid handing an unexpected
    # parameter to the connection function.
    exitOnError = config['opsi'].pop('exitOnError', False)

    if cliArguments.checkOpsi:
        checkConnectionToOpsi(config['opsi'], rootGroup)
        return  # No further processing

    searchQuery = createSearchQuery(
        computerQuery=config['directory'].pop('searchQueryForComputers', None),
        groupQuery=config['directory'].pop('searchQueryForGroups', None),
    )
    elements = readDirectory(mapping=config['mapping'], searchQuery=searchQuery, **config['directory'])
    elements = (convertElementToDict(element) for element in elements)
    if cliArguments.checkDirectory:
        checkConnectionToDirectory(elements)
        return  # No further processing

    writeChanges = config['behaviour'].get('write_changes_to_opsi', False)

    with getSession(**config['opsi']) as opsiSession:
        modulesCheck(opsiSession)
        rootGroupCheck(opsiSession, rootGroup)
        hostIDs = getIdsOfHostsInOpsi(opsiSession)
        opsiClientGroupMapping = getClientGroupMapping(opsiSession)

        actions = convertElements(
            elements, hostIDs, opsiClientGroupMapping,
            opsiSession, config['mapping'], rootGroup, **config['behaviour']
        )
        updateOpsi(opsiSession, actions, writeChanges=writeChanges, exitOnError=exitOnError)

        opsiSession.get(params=json.dumps({"method": "backend_exit", "params": [], "id": 1}))


def parseCommandlineArguments():
    parser = argparse.ArgumentParser(description="opsi directory connector")
    parser.add_argument('--version', action="version", version=__version__)
    parser.add_argument('--log-level', '-l', dest='logLevel',
                        help=(
                            'Sets how much information will be logged. '
                            'From 1 (debug) to 5 (critical)'
                        ),
                        type=int, choices=[1, 2, 3, 4, 5], default=3)
    checks = parser.add_mutually_exclusive_group()
    checks.add_argument('--check-directory', dest="checkDirectory",
                        action="store_true",
                        help=(
                            'Check if a connection to the directory '
                            'can be established and if items will be '
                            'received.')
                        )
    checks.add_argument('--check-opsi', dest="checkOpsi",
                        action="store_true",
                        help=(
                            'Check if a connection to the opsi server '
                            'can be established.')
                        )
    parser.add_argument('--config', help='Path to the config.', required=True)
    return parser.parse_args()


def configureLogger(logger, level):
    """
    Configure the logger.

    The basic level is set to `level`.

    Messages that are in level WARNING, ERROR or CRITICAL will be
    shown in stderr and stdout.
    """
    logger.setLevel(level)

    formatter = logging.Formatter('[%(levelname)s] [%(asctime)s] %(message)s')
    stdoutHandler = logging.StreamHandler(sys.stdout)
    stdoutHandler.setLevel(level)
    stdoutHandler.setFormatter(formatter)
    logger.addHandler(stdoutHandler)


def readConfig(path):
    with open(path, 'r') as configFile:
        return json.load(configFile)


def convertConfig(config):
    """
    Convert the given `config` into a format that allows passing the \
config to our processing functions.
    """
    mappings = {
        "directory": {
            "search_base": "searchBase",
            "search_query_computers": "searchQueryForComputers",
            "search_query_groups": "searchQueryForGroups",
            "connection_options": "connectionOptions",
            "identifying_attribute": "identifyingAttribute",
        },
        "behaviour": {
            "update_existing_clients": "updateExistingClients",
            "prefer_location_from_directory": "preferLocationFromDrectory",
        },
        "opsi": {
            "connection_options": "connectionOptions",
            "exit_on_error": "exitOnError",
        }
    }

    newConfig = {}

    for branch, configs in config.items():
        newConfig[branch] = {}

        for key, value in configs.items():
            try:
                newKey = mappings[branch][key]
                newConfig[branch][newKey] = value
            except KeyError:
                newConfig[branch][key] = value

    removePrefixFromMapping(newConfig, "client", "c")
    removePrefixFromMapping(newConfig, "group", "g")

    for branch in ('directory', 'opsi'):
        try:
            path = newConfig[branch].pop('passwordFile')
            if path:
                with open(path, 'r') as credentials:
                    newConfig[branch]['password'] = credentials.read()
        except KeyError:
            pass

    if 'identifyingAttribute' not in newConfig['directory']:
        newConfig['directory']['identifyingAttribute'] = 'dn'

    if 'group_name' not in newConfig['mapping']:
        newConfig['mapping']['group_name'] = {}

    return newConfig


def modulesCheck(session):
    hostCount = len(getFromOpsi(session, "host_getIdents", ['unicode', {"type": "OpsiClient"}]))
    LOGGER.debug("Clients in opsi: {0}".format(hostCount))
    backendInfo = getFromOpsi(session, "backend_info")
    try:
        modules = backendInfo['modules']
        realmodules = backendInfo.get('realmodules', {})
        if not modules:
            raise ModulesError("No Modules detected.")
        if not modules.get('valid'):
            raise ModulesError('Modules file invalid.')
        elif not modules.get('customer'):
            raise ModulesError('No customer detected in modules file.')
        elif (modules.get('expires', '') != 'never') and (time.mktime(time.strptime(modules.get('expires', '2000-01-01'), "%Y-%m-%d")) - time.time() <= 0):
            raise ModulesError('Modules file expired.')
        elif not modules.get('directory-connector'):
            raise ModulesError('Module directory-connector not available in modules.')

        if realmodules and hostCount:
            countInModules = int(realmodules.get('directory-connector', 0))
            if countInModules and hostCount > countInModules + 50:
                raise ModulesError('Client limit {0} reached. Upgrade your license!'.format(countInModules))
    except Exception as verificationError:
        LOGGER.critical("Missing or invalid licensing detected ({0!s}). Please contact sales@uib.de!".format(verificationError))
        raise verificationError


def removePrefixFromMapping(config, objectType, prefix):
    "Removes the prefix `prefix` from the mapping configued in `config` for `objectType`."
    mapping = config['mapping'][objectType]
    for key in list(mapping.keys()):
        if key.startswith(prefix):
            modifiedKey = key[1:]
            mapping[modifiedKey] = mapping[key]
            del mapping[key]

    config['mapping'][objectType] = mapping


def checkConnectionToOpsi(connectionConfiguration, rootGroup):
    with getSession(**connectionConfiguration) as opsiSession:
        modulesCheck(opsiSession)
        rootGroupCheck(opsiSession, rootGroup)
        backendInfo = getFromOpsi(opsiSession, 'backend_info')
        LOGGER.info("Connected to opsi service v{0}".format(backendInfo['opsiVersion']))


def readDirectory(address, user, password, searchBase, mapping,
                  searchQuery, identifyingAttribute, connectionOptions={}):

    LOGGER.debug("LDAP search query: {!r}".format(searchQuery))
    attr = DEFAULT_ATTRIBUTES.union(getAdAttributes(mapping))
    LOGGER.debug("Adding {!r} as identifying attribute".format(identifyingAttribute))
    attr.add(identifyingAttribute)
    LOGGER.debug("Querying for the following attributes: {0}".format(', '.join(attr)))

    pagedSearch = connectionOptions.pop('paged_search_limit', 0)

    with directoryConnection(address, user=user, password=password, connectionOptions=connectionOptions) as conn:
        if pagedSearch:
            searchResults = getPagedSearchResults(conn, searchBase, searchQuery, attr, pageSize=pagedSearch)
        else:
            searchResults = getSearchResults(conn, searchBase, searchQuery, attr)

        for entry in searchResults:
            LOGGER.debug('yielding element: {!r}'.format(entry))
            yield entry

        LOGGER.debug("Finished returning elements from directory")


def createSearchQuery(computerQuery=None, groupQuery=None):
    """
    Create a search query for our search.

    :raises ValueError: If no queries are given.
    :returntype: str
    """
    groupQuery = None  # resetting this until the feature is implemented.
    if computerQuery and groupQuery:
        return '(|{}{})'.format(computerQuery,
                                groupQuery)
    elif computerQuery:
        return computerQuery
    elif groupQuery:
        return groupQuery
    else:
        raise ValueError("Missing search configuration!")


def getPagedSearchResults(connection, searchBase, searchQuery, attributes, pageSize):
    """
    Return results of the search and use a paged search for searching.
    If a paged search is not possible this falls back to a non-paged search.

    This works as a generator.

    :param connection: The connection to use.
    :type connection: ldap3.Connection
    :param searchBase: Base for the search
    :type searchBase: str
    :param searchQuery: The filter for our search.
    :type searchQuery: str
    :param attributes: The attributes we want.
    :type attributes: [str, ]
    :param pageSize: The number of entries returned per page.
    :type pageSize: int
    """
    if SUBTREE is None:
        LOGGER.warning("Unable to import SUBTREE. Falling back to non-paged-search.")
        entryGenerator = getSearchResults(connection, searchBase, searchQuery, attributes)
    else:
        entryGenerator = connection.extend.standard.paged_search(
            search_base=searchBase,
            search_filter=searchQuery,
            search_scope=SUBTREE,
            attributes=list(attributes),
            paged_size=pageSize,
            generator=True
        )
        LOGGER.debug('Searching for elements in directory with paged search...')

    LOGGER.info('Connected to directory')
    for entry in entryGenerator:
        LOGGER.debug('Found element: {!r}'.format(entry))
        yield entry


def getSearchResults(connection, searchBase, searchQuery, attributes):
    """
    Return results of the search.

    This works as a generator.

    :param connection: The connection to use.
    :type connection: ldap3.Connection
    :param searchBase: Base for the search
    :type searchBase: str
    :param searchQuery: The filter for our search.
    :type searchQuery: str
    :param attributes: The attributes we want.
    :type attributes: [str, ]
    """
    connection.search(searchBase, searchQuery, attributes=list(attributes))
    LOGGER.info('Connected to directory')

    LOGGER.debug('Searching for elements in directory...')
    try:
        for entry in connection.entries:
            LOGGER.debug('Found element: {!r}'.format(entry))
            yield entry
    except AttributeError as attrError:
        LOGGER.debug("Caught attribute error: {0}".format(attrError))
        LOGGER.debug('result: {0}'.format(connection.result))

        LOGGER.debug("Trying fallback for old versions of ldap3...")
        try:
            for entry in connection.response:
                LOGGER.debug('Found element: {!r}'.format(entry))
                yield entry
        except AttributeError as fallBackAttrError:
            LOGGER.debug("Fallback failed: {0!r}".format(fallBackAttrError))

            raise attrError  # We want the original error


def convertElementToDict(element):
    """
    Converts the given `element` to a dict.

    This aims at working with different versions of LDAP3.

    :type element: dict or an element returned from an LDAP search.
    :returntype: dict
    """
    if isinstance(element, dict):
        return element

    try:
        elemDict = json.loads(element.entry_to_json())
    except AttributeError as attErr:
        LOGGER.debug("Unable to get {0!r} as dict: {1!r}".format(element, attErr))
        LOGGER.debug("The element offers the following: {0}".format(dir(element)))

        try:
            elemDict = element.entry_attributes_as_dict()
        except AttributeError as attrErr:
            LOGGER.debug("Failed to call entry_attributes_as_dict: {!r}".format(attrErr))

            elemDict = element.entry_get_attributes_dict()

    LOGGER.debug("elemDict is: {0!r}".format(elemDict))

    if 'dn' not in elemDict:
        LOGGER.debug("Missing DN. Trying to read...")
        try:
            elemDict['dn'] = element.entry_get_dn()
        except AttributeError as attrErr:
            LOGGER.critical("Unable to get DN for element!")
            raise attrErr

    return elemDict


def checkConnectionToDirectory(elements):
    index = 0
    try:
        for index, element in enumerate(elements, start=1):
            LOGGER.debug("Picked up element #{0}: {1}".format(index, element))
        LOGGER.info("The current settings matched {0} elements.".format(index))
    except Exception as connectionError:
        LOGGER.critical("Connecting to the directory failed: {0}".format(connectionError))

    if not index:
        LOGGER.warning("No elements found. Does your search criteria fit?")


def getSession(address, username, password, connectionOptions):
    session = requests.Session()
    session.auth = (username, password)
    session.headers.update({"user-agent": str(USERAGENT)})

    opsiEndpoint = getOpsiEndpointAddress(address)

    # We prepare the get method so that the settings for
    # certificate verification and the appropriate address
    # are always given to avoid passing this to any call.
    verifyCertificate = connectionOptions.get('verify_certificate', True)
    LOGGER.debug("Verify certificate? {}".format(verifyCertificate))
    session.get = partial(session.get, opsiEndpoint, verify=verifyCertificate)

    return session


def getOpsiEndpointAddress(address):
    address = address.strip()
    if not address.endswith('/rpc'):
        if not address.endswith('/'):
            address = address + '/'

        address = address + 'rpc'

    return address


def getAdAttributes(mapping):
    attr = set()
    for objectType in ('client', 'group'):
        attributeMapping = mapping[objectType]

        for opsiAttribute, directoryAttribute in attributeMapping.items():
            if directoryAttribute:
                attr.add(directoryAttribute)
            else:
                LOGGER.warning("No value given for opsi {} attribute {!r}".format(objectType, opsiAttribute))

    return attr


@contextmanager
def directoryConnection(address, user, password, connectionOptions):
    destination = getDestination(address)

    with Connection(destination, user=user, password=password) as connection:
        if connectionOptions.get('start_tls', True):
            connection.start_tls()

        if not connection.bind():
            raise RuntimeError("Connection failed: {0}".format(connection.result))

        yield connection


def getDestination(address):
    destination = address

    if '://' not in destination:
        LOGGER.warning("Directory address misses protocol, adding 'ldap://'.")
        destination = 'ldap://' + destination

    addrInfo = urlsplit(destination)
    LOGGER.debug("Splitted {0!r} to {1!r}".format(destination, addrInfo))
    if addrInfo.port:
        return Server(addrInfo.hostname, port=addrInfo.port)

    return destination


def convertElements(elements, hostsInOpsi, opsiClientGroupMappings,
                    session, attributeMapping, rootGroup,
                    updateExistingClients=False,
                    preferLocationFromDrectory=False,
                    identifyingAttribute='dn', **_):

    def isComputer(dn):
        computerDetected = False
        allElements = set(str(dn).split(","))
        for element in allElements:
            key, value = element.split('=', 1)

            if key.lower() == 'cn' and value.lower().startswith('computer'):
                computerDetected = True
                break

        LOGGER.debug("{0} is computer? {1}".format(dn, computerDetected))
        return computerDetected

    def deleteStrObj2Grp(objectToGroup):
        return {
           'method': "objectToGroup_deleteObjects",
           'params': [json.dumps(objectToGroup)],
           'id': 1}

    def deleteObj2Grp(clientname):
        def isFromRootDir(groupname):
            if groupname is None:
                return False
            elif groupname == rootGroup:
                return True

            for grp in groupsInOpsi:
                if groupname == grp["id"]:
                    result = isFromRootDir(grp["parentGroupId"])
                    LOGGER.debug("isFromRootDir parentGroupId {0!r}: {1}".format(grp["parentGroupId"], result))
                    return result

        for clientGroupMapping in opsiClientGroupMappings:
            if clientname == clientGroupMapping["objectId"]:
                groupname = clientGroupMapping["groupId"]
                LOGGER.debug("Found group {1!r} for client {0!r}".format(clientname, groupname))
                if isFromRootDir(groupname):
                    return deleteStrObj2Grp(clientGroupMapping)

    clientAttributes = createClientMapping(attributeMapping['client'])
    groupMapping = createGroupMapping(attributeMapping['group'])
    groupNameReplacements = attributeMapping['group_name']

    groupsInOpsi = getGroupsFromOpsi(session)
    opsiHosts = set(hostsInOpsi)
    for element in elements:
        LOGGER.debug("Next element: {0!r}".format(element))

        try:
            dn = getDNFromElement(element, identifyingAttribute)

            try:
                category = getAttributeFromElement(element, 'objectCategory', flatten=True)
                fullDN = "{0},{1}".format(category, dn)
            except KeyError as kerr:
                LOGGER.debug("Creating full DN failed: {0!r}".format(kerr))
                fullDN = dn

            parentlist = getGrouplist(fullDN)
            parentlist = replaceGroups(parentlist, groupNameReplacements)
            if isComputer(fullDN):
                try:
                    name = getAttributeFromElement(element, clientAttributes.id, flatten=True)
                except KeyError as kerr:
                    LOGGER.debug("{0!r} during reading of the attribute for the client id.".format(kerr))
                    LOGGER.critical("Unable to read the attribute {attr!r} specified for the client ID. Aborting.".format(attr=clientAttributes.id))
                    raise kerr

                if not isFqdn(name):
                    LOGGER.debug("Need to create a fqdn...")
                    fqdn = "{0}.{1}".format(name, getDomain(dn))
                else:
                    fqdn = name

                fqdn = fqdn.lower()  # opsi expected FQDN in lowercase.
                LOGGER.debug("Using {0!r} as fqdn".format(fqdn))

                if fqdn in opsiHosts:
                    LOGGER.info("Host {0!r} exists in opsi".format(fqdn))
                    if updateExistingClients:
                        LOGGER.info("Updating existing host {0!r}".format(fqdn))
                        if preferLocationFromDrectory:
                            LOGGER.debug("Removing existing object2Groups {0!r}".format(fqdn))
                            yield deleteObj2Grp(fqdn)

                        LOGGER.info("Checking groups of host {0!r}".format(fqdn))
                        groupsChanged = False
                        for groupAction in processGroupsOfHost(fqdn, parentlist, rootGroup, groupsInOpsi):
                            yield groupAction
                            groupsChanged = True

                        if groupsChanged:
                            groupsInOpsi = getGroupsFromOpsi(session)

                        yield getClientUpdateAction(session, fqdn, element, clientAttributes._asdict())
                    else:
                        LOGGER.debug("Skipped existing host {!r}".format(element))
                else:
                    LOGGER.info("Host {0!r} does not exist in opsi.".format(fqdn))
                    yield {
                        'method': 'host_createOpsiClient',
                        'params': [
                            fqdn,
                            None,  # HostKey - not setting one
                            convertToStringIfNotNone(element, clientAttributes.description),
                            convertToStringIfNotNone(element, clientAttributes.notes),
                            convertToStringIfNotNone(element, clientAttributes.hardwareAddress),
                            convertToStringIfNotNone(element, clientAttributes.ipAddress),
                            convertToStringIfNotNone(element, clientAttributes.inventoryNumber),
                            convertToStringIfNotNone(element, clientAttributes.oneTimePassword),
                        ],
                        'id': 1}

                    LOGGER.info("Setting groups for new host {0!r}".format(fqdn))
                    groupsChanged = False
                    for groupAction in processGroupsOfHost(fqdn, parentlist, rootGroup, groupsInOpsi):
                        yield groupAction
                        groupsChanged = True

                    if groupsChanged:
                        groupsInOpsi = getGroupsFromOpsi(session)
            else:
                LOGGER.info("Element {0} is not a computer. Syncing currently not supported.".format(element[identifyingAttribute]))
        except LDAPKeyError as error:
            LOGGER.warning("Processing {} failed: {}".format(element, error))
            LOGGER.debug("Full exception: {!r}".format(error))


def getDNFromElement(element, identifyingAttribute='dn'):
    try:
        return element[identifyingAttribute]
    except KeyError as kerr:
        LOGGER.debug("Failed to get identifying name from {0!r} (via attribute {attr!r}): {error!r}".format(element, error=kerr, attr=identifyingAttribute))
        LOGGER.critical("Unable to get the identifying name for the current element!")
        raise kerr


def getGrouplist(dn):
    """
    Gets the groups from a DN.

    As a group every Organisational Unit (OU) qualifies.

    :type dn: str
    :returntype: [str, ]
    :return: groups in the dn with the groupnames being lowercase.
    """
    groups = []
    for element in dn.lower().split(","):
        if element.startswith("ou="):
            group = element.replace("ou=", "", 1).strip()
            groups.append(group)

    LOGGER.debug("Groups {0!r} extracted from {1}".format(groups, dn))
    return groups


def replaceGroups(groups, replacements):
    """
    Apply the `replacements` to `groups`.

    :type groups: list
    :type replacements: Replace every occurance of `key` with `value`.
    :type replacements: {str: str}
    """
    for index, group in enumerate(groups):
        try:
            replacement = replacements[group]
            groups[index] = replacement.lower().strip()
            LOGGER.debug("Replaced group {!r} with {!r}".format(group, replacement))
        except KeyError:
            pass

    return groups


def getAttributeFromElement(element, attribute, flatten=True):
    """
    Return the value for `attribute` in element.

    This is designed to work with different versions of dicts created
    from different versions of ldap3.

    :param element: The element to search the attribute for.
    :type element: dict
    :param attribute: The attribute to search for.
    :type attribute: str
    :param flatten: If the attributes holds many values this selects the first value to be returned.
    :type flatten: bool
    :raises KeyError: If no matching attribute is found.
    """
    try:
        attr = element['attributes'][attribute]
    except KeyError as kerr:
        LOGGER.debug("Trying direct access to {0!r} after accessing in 'attributes' failed with {1!r}".format(attribute, kerr))
        attr = element[attribute]

    if flatten and isinstance(attr, list):
        return attr[0]

    return attr


def isFqdn(name):
    return name.count('.') >= 2


def getDomain(dn):
    """
    Get the domain from the given dn.

    :type dn: str
    :return: the domain from the DN in format `domain.tld`
    :returntype: str
    """
    domainComponents = ("DC=", "dc=")
    domainParts = []
    for element in dn.split(","):
        for domainComponent in domainComponents:
            if element.startswith(domainComponent):
                domainParts.append(element.replace(domainComponent, ""))

    domain = ".".join(domainParts)
    LOGGER.debug("domain {0!r} created from {1}".format(domain, dn))
    return domain


def getParent(element, groups, rootGroup):
    """
    Searching for the parent of `element` in `groups`.
    Case is ignored during the search.

    If no parent is found then `rootGroup` will be returned.

    :type element: str
    :type groups: [str, ]
    :type rootGroup: str
    """
    LOGGER.debug('Looking for {0!r} in groups {1!r}'.format(element, groups))

    normalizedGroups = [group.lower() for group in groups]

    try:
        positionOfCurrent = normalizedGroups.index(element.lower())
        groupLen = len(groups)
        if positionOfCurrent <= (groupLen - 1):
            if (positionOfCurrent + 1) <= (groupLen - 1):
                parent = groups[positionOfCurrent + 1]
                LOGGER.debug("Returning parent {0!r}".format(parent))
                return parent
    except ValueError:
        LOGGER.debug("{!r} not in groups.".format(element))

    LOGGER.debug("Returning root {!r}".format(rootGroup))
    return rootGroup


def createClientMapping(clientAttributeMapping):
    return ClientMapping(
        clientAttributeMapping["id"],
        clientAttributeMapping.get("description", ""),
        clientAttributeMapping.get("notes", ""),
        clientAttributeMapping.get("hardwareAddress", ""),
        clientAttributeMapping.get("ipAddress", ""),
        clientAttributeMapping.get("inventoryNumber", ""),
        clientAttributeMapping.get("oneTimePassword", "")
    )


def createGroupMapping(groupAttributeMapping):
    return GroupMapping(
        groupAttributeMapping["id"],
        groupAttributeMapping.get("description", ""),
        groupAttributeMapping.get("notes", ""),
    )


def processGroupsOfHost(clientId, parentGroups, rootGroup, groupsInOpsi):
    try:
        if not parentGroups:
            raise IndexError("No parents for {}.".format(clientId))

        for action in processParentList(parentGroups, rootGroup, groupsInOpsi):
            yield action

        parent = parentGroups[0]
        LOGGER.debug("Parent for {!r} is {!r}".format(clientId, parent))
        yield createObj2Grp(parent, clientId)
    except IndexError:
        LOGGER.debug("No groups for {0!r} found.".format(clientId))


def processParentList(parentList, rootGroup, groupsInOpsi):
    """
    Make sure that the groups in `parentList` exist in opsi.

    :param parentList: List of group parents - usually a _path_.
    :type parentList: [str, ]
    :param rootGroup: The root group under which new groups are to be created.
    :type rootGroup: str
    :param groupsInOpsi: Groups already existing in opsi.
    :type groupsInOpsi: [dict, ]
    """
    def createGroup(groupName):
        LOGGER.info("Creating missing group {0!r}".format(groupName))
        action = createHostGroupCommand(groupName, parentGroupId)
        groupsInOpsi.append(getGroupDefinition(action))
        return action

    def getGroupDefinition(jsonRpcCall):
        definition = jsonRpcCall['params'][0]
        definition['id'] = definition['id'].lower()  # opsi requirement
        return definition

    reversedGroupList = parentList[::-1]

    for groupName in reversedGroupList:  # start at the group near the root.
        parentGroupId = getParent(groupName, parentList, rootGroup)
        LOGGER.debug("Found parent {0!r}".format(parentGroupId))

        if not groupExistsInOpsi(groupName, groupsInOpsi):
            yield createGroup(groupName)
        else:
            if hasGroupRightParent(groupName, parentGroupId, groupsInOpsi):
                LOGGER.debug("Group {0!r} exists in opsi and is at the correct location.".format(groupName))
                continue

            originalGroupName = groupName

            while groupExistsInOpsi(groupName, groupsInOpsi):
                if hasGroupRightParent(groupName, parentGroupId, groupsInOpsi):
                    break  # this is the correct group
                else:
                    groupName = groupName + NAME_ADDITION

            if groupName != originalGroupName:
                LOGGER.debug("Need to correct name of parent...")
                parentList = listReplace(parentList, originalGroupName, groupName)
                LOGGER.debug("Parents are now: {0!r}".format(parentList))

            if not groupExistsInOpsi(groupName, groupsInOpsi):
                yield createGroup(groupName)


def groupExistsInOpsi(name, existingGroups):
    """
    Check if a group with `name` does already exist in opsi.

    :param name: Name of the group to check for.
    :type name: str
    :param existingGroups: An iterable containing dicts with \
information about the existing groups.
    :type existingGroups: [dict, ]
    """
    lowecaseName = name.lower()
    exists = any(lowecaseName == group['id'] for group in existingGroups)
    LOGGER.debug("Exists group {0!r} in opsi? {1}".format(name, exists))
    return exists


def createHostGroupCommand(name, parentGroupId):
    if not parentGroupId:
        raise ValueError("Missing ID for parent or root group.")

    if name == parentGroupId:
        raise Exception("Groupname is same as parent groupname. Forbidden!")

    return {
        'method': "group_insertObject",
        'params': [{
            "id": name,
            "description": None,  # TODO: possible to insert?
            "notes": None,  # TODO: possible to insert?
            "parentGroupId": parentGroupId,
            "type": "HostGroup"
        }],
        'id': 1}


def hasGroupRightParent(groupId, parentGroupId, existingGroups):
    """
    Checks if the group with id `groupId` and parent `parentGroupId` \
exist in `existingGroups`.

    :type groupId: str
    :type parentGroupId: str
    :type existingGroups: [dict, ]
    :returns: `True` if a match is found. `False` otherwise.
    :rtype: bool
    """
    LOGGER.debug("Checking if {0!r} has parent {1!r}".format(groupId, parentGroupId))
    LOGGER.debug("Existing groups: {0}".format(existingGroups))

    name = groupId.lower()
    for group in iterateHostGroups(existingGroups):
        if group['id'].lower() == name and group['parentGroupId'] is not None:
            parentMatches = group['parentGroupId'].lower() == parentGroupId.lower()
            LOGGER.debug("Parent matches? {0}".format(parentMatches))

            # Group ID is unique and we can stop searching now.
            return parentMatches

    LOGGER.debug("No match of group {0} with parent {1} found.".format(groupId, parentGroupId))
    return False


def iterateHostGroups(groups):
    """
    Iterate over all groups that are of type `HostGroup`.

    :param groups:
    :type groups: [dict, ]
    :rtype: dict
    """
    for group in groups:
        if group['type'] == 'HostGroup':
            yield group


def listReplace(llist, find, replaceWith):
    for index, item in enumerate(llist):
        if item == find:
            LOGGER.debug('replacing {0!r} with {1!r}'.format(llist[index], replaceWith))
            llist[index] = replaceWith

    return llist


def createObj2Grp(groupname, clientname):
    return {
        'method': "objectToGroup_createObjects",
        'params': [{
            "groupId": groupname,
            "objectId": clientname,
            "groupType": "HostGroup",
            "type": "ObjectToGroup"
        }],
        'id': 1
    }


def convertToStringIfNotNone(element, attribute):
    """
    Read the value for `attribute` from `element`.

    If the value is not `None`, then it will be explicitely converted to str.
    If there is no attribute found `None` will be returned.
    If no attribute is given `None` will be returned.

    :type attribute: str
    :param attribute: The attribute to search for.
    :returntype: str or None
    """
    if not attribute:
        return None

    try:
        attr = getAttributeFromElement(element, attribute, flatten=True)
        if attr is not None:
            return str(attr)
    except IndexError as ierr:
        LOGGER.debug("No {attr!r} in {elem!r} probably not filled. Received {err!r}".format(attr=attribute, elem=element, err=ierr))
    except KeyError as kerr:
        LOGGER.debug("No {attr!r} in {elem!r}? Received {err!r}".format(attr=attribute, elem=element, err=kerr))

    return None


def getClientUpdateAction(session, clientId, element, clientAttributeMapping):
    opsiHost = getHostFromOpsi(session, clientId)

    updateNecessary = False
    for opsiAttribute, dirAttribute in clientAttributeMapping.items():
        if not dirAttribute or opsiAttribute == 'id':
            # We limit the updating to the attributes that have a
            # mapping and we also do not want override the id because
            # changing the ID would result in an different opsi object.
            continue

        valueInOpsi = opsiHost[opsiAttribute]
        valueInDirectory = convertToStringIfNotNone(element, clientAttributeMapping[opsiAttribute])

        LOGGER.debug("Comparing {0!r} (opsi) and {1!r} (directory)".format(valueInOpsi, valueInDirectory))
        if areValuesDifferent(valueInOpsi, valueInDirectory):
            LOGGER.debug("Difference detected. Updating {attr} of our opsi host.".format(attr=opsiAttribute))
            opsiHost[opsiAttribute] = valueInDirectory
            updateNecessary = True

    if updateNecessary:
        LOGGER.info("Updating attributes of host {0!r}".format(clientId))
        return {
            'method': 'host_insertObject',  # use insert to make sure that values with None are updated aswell
            'params': [opsiHost],
            'id': 1
        }


def areValuesDifferent(firstValue, secondValue):
    """
    Returns `True` if the values differ.

    This will not compare two falsy values, i.e. '' and None.
    """
    return any((firstValue, secondValue)) and firstValue != secondValue


def getHostFromOpsi(session, clientId):
    hosts = getFromOpsi(session, 'host_getHashes', [None, {"id": clientId}])
    return hosts[0]


def updateOpsi(session, actions, writeChanges=False, exitOnError=False):
    processableActions = (action for action in actions if action is not None)

    actionCounter = 0
    for actionCounter, action in enumerate(processableActions, start=1):
        LOGGER.info('Processing {!r}'.format(action['method']))
        LOGGER.debug('Complete call: {!r}'.format(action))
        if writeChanges:
            response = session.get(params=json.dumps(action))
            content = response.json()

            if content['error']:
                errorMessage = "Call to opsi failed: {class} - {message}".format(**content['error'])
                if exitOnError:
                    raise RuntimeError(errorMessage)
                else:
                    LOGGER.error("ERROR: {}".format(errorMessage))
            LOGGER.debug("Result from call to opsi: {0}".format(content["result"]))
            LOGGER.info('Processing {!r} finished'.format(action['method']))

    LOGGER.debug("Processed {0} action(s) in opsi.".format(actionCounter))


def rootGroupCheck(session, groupName):
    groups = getFromOpsi(session, "group_getHashes", [None, {"id": groupName}])

    if not groups:
        raise ValueError("Group with id {!r} not found in opsi.".format(groupName))


def getIdsOfHostsInOpsi(session):
    return set(getFromOpsi(session, "host_getIdents"))


def getClientGroupMapping(session):
    return getFromOpsi(session, "objectToGroup_getHashes")


def getGroupsFromOpsi(session):
    return getFromOpsi(session, "group_getHashes")


def getFromOpsi(session, method, params=None):
    params = json.dumps({"method": method, "id": 1, "params": params or []})
    response = session.get(params=params)
    content = response.json()
    if content['error']:
        raise RuntimeError("Call to opsi failed: {class} - {message}".format(**content['error']))
    LOGGER.debug("get_from_opsi {0}: {1}".format(method, content["result"]))
    return content['result']


if __name__ == '__main__':
    runConnector()
